/* $Header: /fridge/cvs/cgi.targeted/targeted-log.c,v 1.15 2002/06/30 19:28:40 justins Exp $ */
/*
   
   targeted-log.c             Copyright(c) 2002-1998 Justin David Smith
   justins(at)chaos2.org      http://chaos2.org/
    
   Logging functions
    

   This program is free software; you can redistribute it and/or modify 
   it under the terms of the GNU General Public License as published by 
   the Free Software Foundation, version 2 of the License ONLY.

   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation, 
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <targeted-log.h>
#include <targeted-cgi.h>
#include <targeted-util.h>


static void _log_internal(const config *c, const char *mode, const char *msg) {
/* Internal logging mechanism */

   char buf[0x1000];
   const char *host;
   const char *ip;
   const char *qs;
   struct timeval tv;
   struct tm *tm;
   FILE *f;
   
   gettimeofday(&tv, NULL);
   tm = localtime(&tv.tv_sec);
   sbprintf(buf, sizeof(buf), "%04d.%02d.%02d %02d:%02d:%02d", tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);
   
   host = stra(c->env.remote_host);
   ip   = stra(c->env.remote_addr);
   qs   = stra(c->env.query_string);
   
   if((f = fopen(stra(c->logfile), "a")) == NULL) return;
   fprintf(f, "%s|%s|%s|%s|%s|%s\n", mode, buf, ip, host, qs, msg);
   fclose(f);

}


void targeted_log_request(const config *c) {
/* Log a request */

   assert(c != NULL);
   _log_internal(c, "request", "");

}


void targeted_log_error(const config *c, const char *msg) {
/* Log an error */

   assert(c != NULL);
   assert(msg != NULL);
   _log_internal(c, "error", msg);

}


void targeted_log_error_access(const config *c, int acc) {
/* Log an error message based on access result */

   switch(acc) {
   case FILE_NOT_FOUND:
      targeted_log_error(c, "file not found");
      break;
   case HTACCESS_DENIES:
      targeted_log_error(c, "an htaccess file denies access");
      break;
   case PERM_DENIES:
      targeted_log_error(c, "file permissions deny access");
      break;
   default:
      targeted_log_error(c, "an unknown condition denies access");
      break;
   }

}


void targeted_log_message(const config *c, const char *msg) {
/* Log a message */

   assert(c != NULL);
   assert(msg != NULL);
   _log_internal(c, "message", msg);
   
}


static void _print_safely(const unsigned char *s) {
/* Safe printer for bogus characters */

   while(*s != '\0') {
      if(*s >= 0x20 && *s <= 0x7F) {
         if(!(*s == '<' || *s == '>' || *s == '&')) {
            putchar(*s);
         } else putchar('?');
      } else putchar('?'); 
      ++s;
   }

}


static void _targeted_query_string(stringa *query_string) {
/* Reformat the query string in a form that we can destructure */

   strreplacea(query_string, "&amp;", "&");

}


static void _targeted_view_log(const config *c, int count, bool all) {
/* Display the log page, and show the most recent log entries,
   as indicated by the count argument. */

   char buf[0x1000];
   stringa *val = strnewa(NULL);
   stringa *qs = strnewa(NULL);
   const char *cgi_mode;
   const char *cgi_path;
   const char *p;
   bintree *cgi;
   int size = 0;
   int skip = 0;
   FILE *f;

   /* Print out the page header */
   targeted_begin(c, "Logfile");
   printf("<h1>Logfile - %s</h1>\n", all ? "All Entries" : "Errors");

   if(!targeted_admin_ip(c)) {
      printf("<p class=error>Sorry, you are not authorized to view this page.</p>");
      targeted_log_error(c, "Unauthorized request");
      targeted_end(c);
      return;
   }
   
   printf("<div class=center>\n");
   printf("<form action=\"%s\" method=get><p>\n", stra(c->cgi_uri));
   printf("Show the last&nbsp;<input type=text size=5 name=\"count\" value=\"%d\">&nbsp;", count);
   printf("<input type=radio name=mode value=viewlog %s>&nbsp;entries | <input type=radio name=mode value=viewerror %s>&nbsp;errors</p>\n", all ? "checked" : "", all ? "" : "checked");
   printf("<p><input type=submit value=\"Submit\"><input type=reset value=\"Reset\"></p>\n");
   printf("</form>\n");
   printf("</div>\n");

   /* Determine if we should only show recent entries */
   if(count > 0) {
      if((f = fopen(stra(c->logfile), "r")) == NULL) {
         printf("<p class=error>Unable to open logfile for read.</p>");
         targeted_end(c);
         return;
      }
      while(fgets(buf, sizeof(buf), f) != NULL) ++size;
      fclose(f);
      if(size > count) {
         skip = size - count;
         printf("<p>Skipping %d of %d total entries</p>\n\n", skip, size);
      } /* Skipping anything? */
   } /* Check number of log lines */

   /* Try to open file for read */
   if((f = fopen(stra(c->logfile), "r")) == NULL) {
      printf("<p class=error>Unable to open logfile for read.</p>");
      targeted_end(c);
      return;
   }

   /* Begin writing log entries */
   count = 0;
   printf("<table class=log summary=\"Log entries from server\">\n");
   while(skip > 0 && fgets(buf, sizeof(buf), f) != NULL) --skip;
   while(fgets(buf, sizeof(buf), f) != NULL) {
      ++count;
      getarga(val, buf, 0, '|');
      if(!((strequala(val, "request") || strequala(val, "message")) && !all)) {
         if(strequala(val, "request") || strequala(val, "message")) {
            printf("<tr>");
         } else {
            printf("<tr class=error>");
         }

         /* Print time argument */
         getarga(val, buf, 1, '|');
         printf("<td class=left>");
         if(getnumargsa(val, ' ') == 2) {
            getargaa(qs, val, 0, ' ');
            printf("%s", stra(qs));
            getargaa(qs, val, 1, ' ');
            printf("<br>%s", stra(qs));
         } else {
            printf("%s", stra(val));
         }
         printf("</td>");
         
         /* Hostname and IP */
         getarga(val, buf, 2, '|');
         printf("<td>%s<br>", stra(val));
         getarga(val, buf, 3, '|');
         putslenn(stra(val), 22);
         printf("</td>");
         
         /* Handle the damn query string */
         getarga(val, buf, 4, '|');
         _targeted_query_string(val);
         cgi = parse_query_args(stra(val));
         cgi_mode = lookup_cgi_argument(cgi, "mode", "");
         cgi_path = lookup_cgi_argument(cgi, "path", "");
         if(*cgi_mode != '\0') {
            /* New-form URL */
            printf("<td>");
            if(strlenn(cgi_mode) > 34) p = cgi_mode + strlenn(cgi_mode) - 34;
            else p = cgi_mode;
            _print_safely(p);
            if(strequal(cgi_mode, "image")) {
               int cgi_width = atoi(lookup_cgi_argument(cgi, "width", "0"));
               int cgi_height = atoi(lookup_cgi_argument(cgi, "height", "0"));
               printf(" (%dx%d)", cgi_width, cgi_height);
            }
            printf("<br>");
            if(strlenn(cgi_path) > 34) p = cgi_path + strlenn(cgi_path) - 34;
            else p = cgi_path;
            _print_safely(p);
            printf("</td>");
         } else {
            /* Old-form URL; use the old parser */
            getarga(val, buf, 4, '|');
            if(getnumargsa(val, '=') == 2) {
               getargaa(qs, val, 0, '=');
               if(strlena(qs) > 34) p = stra(qs) + strlena(qs) - 34;
               else                 p = stra(qs);
               printf("<td>");
               _print_safely(p);
               getargaa(qs, val, 1, '=');
               if(strlena(qs) > 34) p = stra(qs) + strlena(qs) - 34;
               else                 p = stra(qs);
               printf("<br>");
               _print_safely(p);
               printf("</td>");
            } else {
               printf("<td>%s</td>", stra(val));
            }
         }
         bintree_release(&cgi);

         getarga(val, buf, 5, '|');
         printf("<td class=right>%s</td>", stra(val));
         printf("</tr>\n");
      } /* Skip this record? */
   } /* Iterate over all records */
   printf("</table>\n\n");
   fclose(f);

   /* Write footer */   
   printf("<p>%d log messages were displayed.</p><hr>\n", count);
   targeted_end(c);
   strreleasea(&val);
   strreleasea(&qs);

}


void targeted_view_log_qstr(const config *c, const char *data) {
/* Display the log (old interface) */

   assert(c != NULL);
   assert(data != NULL);
   
   /* Display the log page with the specified count */
   _targeted_view_log(c, atoi(data), true);

}


void targeted_view_log(const config *c) {
/* Display the log */

   int count;

   assert(c != NULL);
   
   /* Lookup the count argument, and display appropriate page */
   count = atoi(lookup_cgi_argument(c->env.query_args, "count", "1000"));
   _targeted_view_log(c, count, true);

}


void targeted_view_error(const config *c) {
/* Display the errors in the log */

   int count;

   assert(c != NULL);
   
   /* Lookup the count argument, and display appropriate page */
   count = atoi(lookup_cgi_argument(c->env.query_args, "count", "0"));
   _targeted_view_log(c, count, false);

}


